/*
 * netcap-advanced.h - Advanced capability analysis header file
 * Copyright (c) 2026 Steve Grubb
 * All Rights Reserved.
 *
 * This software may be freely redistributed and/or modified under the
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING. If not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor
 * Boston, MA 02110-1335, USA.
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 */

#ifndef NETCAP_ADVANCED_H
#define NETCAP_ADVANCED_H

#include "gcc-attributes.h"

struct netcap_opts {
	int advanced;
	int list_interfaces;
	int json;
	int no_color;
	const char *interface;
};

int netcap_advanced_main(const struct netcap_opts *opts)
	__attr_access ((__read_only__, 1))
	__wur;

#endif
